/*
  Copyright (C) 2002-2017 CERN for the benefit of the ATLAS collaboration
*/

// Class: ReadMLP
// Automatically generated by MethodBase::MakeClass
//

#include <vector>
#include <cmath>
#include <string>
#include <iostream>

#ifndef IClassifierReader__def
#define IClassifierReader__def

class IClassifierReader {

 public:

   // constructor
   IClassifierReader() {}
   virtual ~IClassifierReader() {}

   // return classifier response
   virtual double GetMvaValue( const std::vector<double>& inputValues ) const = 0;
};

#endif

class ReadMLP : public IClassifierReader {

 public:

   // constructor
   ReadMLP( std::vector<std::string>& theInputVars ) 
      : IClassifierReader(),
        fClassName( "ReadMLP" ),
        fStatusIsClean( true ),
        fNvars( 6 ),
        fIsNormalised( true )
   {      
      // the training input variables
      const char* inputVars[] = { "ptj", "egcore", "hadcore", "leadpt", "leadptchg", "highm5pt" };

      // sanity checks
      if (theInputVars.size() <= 0) {
         std::cout << "Problem in class \"" << fClassName << "\": empty input vector" << std::endl;
         fStatusIsClean = false;
      }

      if (theInputVars.size() != fNvars) {
         std::cout << "Problem in class \"" << fClassName << "\": mismatch in number of input values: "
                   << theInputVars.size() << " != " << fNvars << std::endl;
         fStatusIsClean = false;
      }

      // validate input variables
      for (size_t ivar = 0; ivar < theInputVars.size(); ivar++) {
         if (theInputVars[ivar] != inputVars[ivar]) {
            std::cout << "Problem in class \"" << fClassName << "\": mismatch in input variable names" << std::endl
                      << " for variable [" << ivar << "]: " << theInputVars[ivar].c_str() << " != " << inputVars[ivar] << std::endl;
            fStatusIsClean = false;
         }
      }

      // initialize min and max vectors (for normalisation)
      fVmin[0] = 17.0000610351562;
      fVmax[0] = 99.9725875854492;
      fVmin[1] = 0;
      fVmax[1] = 24.9998874664307;
      fVmin[2] = 0;
      fVmax[2] = 98.9895172119141;
      fVmin[3] = 1.05319571495056;
      fVmax[3] = 98.9895172119141;
      fVmin[4] = 0;
      fVmax[4] = 98.9895172119141;
      fVmin[5] = 0;
      fVmax[5] = 117.28783416748;

      // initialize input variable types
      fType[0] = 'F';
      fType[1] = 'F';
      fType[2] = 'F';
      fType[3] = 'F';
      fType[4] = 'F';
      fType[5] = 'F';

      // initialize constants
      Initialize();

   }

   // destructor
   virtual ~ReadMLP() {
      Clear(); // method-specific
   }

   // the classifier response
   // "inputValues" is a vector of input values in the same order as the 
   // variables given to the constructor
   double GetMvaValue( const std::vector<double>& inputValues ) const;

 private:

   // method-specific destructor
   void Clear();

   // common member variables
   const char* fClassName;
   bool fStatusIsClean;

   const size_t fNvars;
   size_t GetNvar()           const { return fNvars; }
   char   GetType( int ivar ) const { return fType[ivar]; }

   // normalisation of input variables
   const bool fIsNormalised;
   bool IsNormalised() const { return fIsNormalised; }
   double fVmin[6];
   double fVmax[6];
   double NormVariable( double x, double xmin, double xmax ) const {
      // normalise to output range: [-1, 1]
      return 2*(x - xmin)/(xmax - xmin) - 1.0;
   }

   // type of input variable: 'F' or 'I'
   char   fType[6];

   // initialize internal variables
   void Initialize();
   double GetMvaValue__( const std::vector<double>& inputValues ) const;

   // private members (method specific)

   double ActivationFnc(double x) const;

   int fLayers;
   int fLayerSize[4];
   double fWeightMatrix0to1[12][7];   // weight matrix from layer 0 to 1
   double fWeightMatrix1to2[7][12];   // weight matrix from layer 1 to 2
   double fWeightMatrix2to3[1][7];   // weight matrix from layer 2 to 3

   double * fWeights[4];
};

inline void ReadMLP::Initialize()
{
   // build network structure
   fLayers = 4;
   fLayerSize[0] = 7; fWeights[0] = new double[7]; 
   fLayerSize[1] = 12; fWeights[1] = new double[12]; 
   fLayerSize[2] = 7; fWeights[2] = new double[7]; 
   fLayerSize[3] = 1; fWeights[3] = new double[1]; 
   // weight matrix from layer 0 to 1
   fWeightMatrix0to1[0][0] = -1.35863616970102;
   fWeightMatrix0to1[1][0] = 0.355267825532901;
   fWeightMatrix0to1[2][0] = 1.11968100770938;
   fWeightMatrix0to1[3][0] = 1.63476135793075;
   fWeightMatrix0to1[4][0] = -0.881401114537419;
   fWeightMatrix0to1[5][0] = 1.33099464036524;
   fWeightMatrix0to1[6][0] = -0.687271667189041;
   fWeightMatrix0to1[7][0] = 1.35077317859302;
   fWeightMatrix0to1[8][0] = 0.294200995507164;
   fWeightMatrix0to1[9][0] = -1.48130303337079;
   fWeightMatrix0to1[10][0] = -0.0909154859636571;
   fWeightMatrix0to1[0][1] = 0.959810230401663;
   fWeightMatrix0to1[1][1] = -1.06324092946709;
   fWeightMatrix0to1[2][1] = -5.76336229163837;
   fWeightMatrix0to1[3][1] = -2.07175985540646;
   fWeightMatrix0to1[4][1] = 0.493743211808833;
   fWeightMatrix0to1[5][1] = -1.2891265974912;
   fWeightMatrix0to1[6][1] = 4.24598829927305;
   fWeightMatrix0to1[7][1] = 0.0164430982953178;
   fWeightMatrix0to1[8][1] = 5.48097729924405;
   fWeightMatrix0to1[9][1] = 0.754737219267636;
   fWeightMatrix0to1[10][1] = -0.946890453845568;
   fWeightMatrix0to1[0][2] = 1.18093866713897;
   fWeightMatrix0to1[1][2] = 1.80396001699007;
   fWeightMatrix0to1[2][2] = 1.52717646774325;
   fWeightMatrix0to1[3][2] = 2.33662747894103;
   fWeightMatrix0to1[4][2] = 2.45111772886874;
   fWeightMatrix0to1[5][2] = -2.52604675121866;
   fWeightMatrix0to1[6][2] = -2.08899522797482;
   fWeightMatrix0to1[7][2] = 1.05120556788023;
   fWeightMatrix0to1[8][2] = 5.11881886126749;
   fWeightMatrix0to1[9][2] = 1.38566440548325;
   fWeightMatrix0to1[10][2] = -2.53948815280012;
   fWeightMatrix0to1[0][3] = -3.83920957236234;
   fWeightMatrix0to1[1][3] = -5.43131768659034;
   fWeightMatrix0to1[2][3] = -1.59900627268632;
   fWeightMatrix0to1[3][3] = -6.80972345999108;
   fWeightMatrix0to1[4][3] = 1.55241569423265;
   fWeightMatrix0to1[5][3] = -6.14202954282145;
   fWeightMatrix0to1[6][3] = 5.36065374252205;
   fWeightMatrix0to1[7][3] = 0.0302912087482007;
   fWeightMatrix0to1[8][3] = -2.20277578705225;
   fWeightMatrix0to1[9][3] = 1.76175714498424;
   fWeightMatrix0to1[10][3] = -2.72120334473083;
   fWeightMatrix0to1[0][4] = -0.172322441419825;
   fWeightMatrix0to1[1][4] = -5.13807660097258;
   fWeightMatrix0to1[2][4] = 0.257800572194352;
   fWeightMatrix0to1[3][4] = -2.38940275959441;
   fWeightMatrix0to1[4][4] = 1.66361068160208;
   fWeightMatrix0to1[5][4] = 3.85940163827037;
   fWeightMatrix0to1[6][4] = -2.68691447083723;
   fWeightMatrix0to1[7][4] = 0.420439486625712;
   fWeightMatrix0to1[8][4] = 0.792164705629557;
   fWeightMatrix0to1[9][4] = 1.44406793932154;
   fWeightMatrix0to1[10][4] = 0.482090274915509;
   fWeightMatrix0to1[0][5] = 2.90212135134892;
   fWeightMatrix0to1[1][5] = -0.467329773679144;
   fWeightMatrix0to1[2][5] = -3.44828629211672;
   fWeightMatrix0to1[3][5] = 2.37478884166659;
   fWeightMatrix0to1[4][5] = 1.93095546957089;
   fWeightMatrix0to1[5][5] = -1.81539917249789;
   fWeightMatrix0to1[6][5] = -0.623607995854637;
   fWeightMatrix0to1[7][5] = -0.285335872428147;
   fWeightMatrix0to1[8][5] = -0.803888962558799;
   fWeightMatrix0to1[9][5] = 1.62681417674909;
   fWeightMatrix0to1[10][5] = -7.73343650169979;
   fWeightMatrix0to1[0][6] = 0.196297181875605;
   fWeightMatrix0to1[1][6] = -8.09991145420649;
   fWeightMatrix0to1[2][6] = 2.58514246487067;
   fWeightMatrix0to1[3][6] = -3.16512536957876;
   fWeightMatrix0to1[4][6] = -1.1087033178518;
   fWeightMatrix0to1[5][6] = -3.61844135337377;
   fWeightMatrix0to1[6][6] = -2.17850221642457;
   fWeightMatrix0to1[7][6] = -0.148588223967372;
   fWeightMatrix0to1[8][6] = 1.4548620597591;
   fWeightMatrix0to1[9][6] = -0.428426117424227;
   fWeightMatrix0to1[10][6] = -9.16043701262521;
   // weight matrix from layer 1 to 2
   fWeightMatrix1to2[0][0] = -2.01756746620757;
   fWeightMatrix1to2[1][0] = 1.93315985477623;
   fWeightMatrix1to2[2][0] = 0.586376191536926;
   fWeightMatrix1to2[3][0] = 0.450462742415433;
   fWeightMatrix1to2[4][0] = -0.156965912396646;
   fWeightMatrix1to2[5][0] = 1.56169236818918;
   fWeightMatrix1to2[0][1] = -1.19154563335079;
   fWeightMatrix1to2[1][1] = 0.104425369302282;
   fWeightMatrix1to2[2][1] = 1.11215034967172;
   fWeightMatrix1to2[3][1] = 1.04344037327303;
   fWeightMatrix1to2[4][1] = -1.36111194916732;
   fWeightMatrix1to2[5][1] = -0.5934411867395;
   fWeightMatrix1to2[0][2] = -0.597792885746589;
   fWeightMatrix1to2[1][2] = 0.119601133331117;
   fWeightMatrix1to2[2][2] = 0.0249489435301289;
   fWeightMatrix1to2[3][2] = 0.824322960207608;
   fWeightMatrix1to2[4][2] = -0.652108369211788;
   fWeightMatrix1to2[5][2] = 0.393378677337595;
   fWeightMatrix1to2[0][3] = 0.136847670302133;
   fWeightMatrix1to2[1][3] = -0.641169720832402;
   fWeightMatrix1to2[2][3] = 0.164211018320105;
   fWeightMatrix1to2[3][3] = 0.225216141182485;
   fWeightMatrix1to2[4][3] = -1.82521417318264;
   fWeightMatrix1to2[5][3] = 0.104678584514443;
   fWeightMatrix1to2[0][4] = -0.654253743469733;
   fWeightMatrix1to2[1][4] = -0.737504206242956;
   fWeightMatrix1to2[2][4] = 1.7325685102728;
   fWeightMatrix1to2[3][4] = -0.750263029745144;
   fWeightMatrix1to2[4][4] = 1.68831003559797;
   fWeightMatrix1to2[5][4] = -0.512374593257449;
   fWeightMatrix1to2[0][5] = -1.44137324894219;
   fWeightMatrix1to2[1][5] = 0.386472104891384;
   fWeightMatrix1to2[2][5] = 0.178417459501999;
   fWeightMatrix1to2[3][5] = 0.217291303120707;
   fWeightMatrix1to2[4][5] = -0.524031976509008;
   fWeightMatrix1to2[5][5] = -0.642787303821237;
   fWeightMatrix1to2[0][6] = -0.173391726912647;
   fWeightMatrix1to2[1][6] = -1.30728304244028;
   fWeightMatrix1to2[2][6] = 0.524116377117486;
   fWeightMatrix1to2[3][6] = 0.647137920154457;
   fWeightMatrix1to2[4][6] = -1.94923371512367;
   fWeightMatrix1to2[5][6] = -0.129522211043097;
   fWeightMatrix1to2[0][7] = -0.620232149919463;
   fWeightMatrix1to2[1][7] = 1.23266478704893;
   fWeightMatrix1to2[2][7] = 0.466377470403172;
   fWeightMatrix1to2[3][7] = 1.28246812160841;
   fWeightMatrix1to2[4][7] = -0.396682174654263;
   fWeightMatrix1to2[5][7] = -0.371792672915382;
   fWeightMatrix1to2[0][8] = -0.395261695851815;
   fWeightMatrix1to2[1][8] = -3.06158080081136;
   fWeightMatrix1to2[2][8] = -0.399173564655177;
   fWeightMatrix1to2[3][8] = -0.509866673983801;
   fWeightMatrix1to2[4][8] = 0.132519578949491;
   fWeightMatrix1to2[5][8] = -0.113121939052194;
   fWeightMatrix1to2[0][9] = -1.05589933703286;
   fWeightMatrix1to2[1][9] = -0.348995400582059;
   fWeightMatrix1to2[2][9] = -1.06947911973433;
   fWeightMatrix1to2[3][9] = 0.861430291723568;
   fWeightMatrix1to2[4][9] = 1.11130313790832;
   fWeightMatrix1to2[5][9] = 2.01075385787491;
   fWeightMatrix1to2[0][10] = -1.38574320506867;
   fWeightMatrix1to2[1][10] = 0.297335310833319;
   fWeightMatrix1to2[2][10] = 0.425399695433244;
   fWeightMatrix1to2[3][10] = 0.149270570274458;
   fWeightMatrix1to2[4][10] = -1.52148682746593;
   fWeightMatrix1to2[5][10] = -1.29295042597047;
   fWeightMatrix1to2[0][11] = -3.92413006501098;
   fWeightMatrix1to2[1][11] = 1.53770315964507;
   fWeightMatrix1to2[2][11] = 2.32510532022512;
   fWeightMatrix1to2[3][11] = 2.31673648899074;
   fWeightMatrix1to2[4][11] = -0.936293946316135;
   fWeightMatrix1to2[5][11] = -2.8180691729656;
   // weight matrix from layer 2 to 3
   fWeightMatrix2to3[0][0] = 0.236812517884415;
   fWeightMatrix2to3[0][1] = -0.310842649848729;
   fWeightMatrix2to3[0][2] = -0.17516343751807;
   fWeightMatrix2to3[0][3] = -0.204522042676872;
   fWeightMatrix2to3[0][4] = 0.212815798140234;
   fWeightMatrix2to3[0][5] = 0.259659255800626;
   fWeightMatrix2to3[0][6] = 0.399846765309018;
}

inline double ReadMLP::GetMvaValue__( const std::vector<double>& inputValues ) const
{
   if (inputValues.size() != (unsigned int)fLayerSize[0]-1) {
      std::cout << "Input vector needs to be of size " << fLayerSize[0]-1 << std::endl;
      return 0;
   }

   for (int l=0; l<fLayers; l++)
      for (int i=0; i<fLayerSize[l]; i++) fWeights[l][i]=0;

   for (int l=0; l<fLayers-1; l++)
      fWeights[l][fLayerSize[l]-1]=1;

   for (int i=0; i<fLayerSize[0]-1; i++)
      fWeights[0][i]=inputValues[i];

   // layer 0 to 1
   for (int o=0; o<fLayerSize[1]-1; o++) {
      for (int i=0; i<fLayerSize[0]; i++) {
         double inputVal = fWeightMatrix0to1[o][i] * fWeights[0][i];
         fWeights[1][o] += inputVal;
      }
      fWeights[1][o] = ActivationFnc(fWeights[1][o]);
   }
   // layer 1 to 2
   for (int o=0; o<fLayerSize[2]-1; o++) {
      for (int i=0; i<fLayerSize[1]; i++) {
         double inputVal = fWeightMatrix1to2[o][i] * fWeights[1][i];
         fWeights[2][o] += inputVal;
      }
      fWeights[2][o] = ActivationFnc(fWeights[2][o]);
   }
   // layer 2 to 3
   for (int o=0; o<fLayerSize[3]; o++) {
      for (int i=0; i<fLayerSize[2]; i++) {
         double inputVal = fWeightMatrix2to3[o][i] * fWeights[2][i];
         fWeights[3][o] += inputVal;
      }
   }

   return fWeights[3][0];
}

double ReadMLP::ActivationFnc(double x) const
{
   // hyperbolic tan
   return tanh(x);
}
 
   
// Clean up
inline void ReadMLP::Clear() 
{
   // nothing to clear
}
inline double ReadMLP::GetMvaValue( const std::vector<double>& inputValues ) const
{
   // classifier response value
   double retval = 0;

   // classifier response, sanity check first
   if (!fStatusIsClean) {
      std::cout << "Problem in class \"" << fClassName << "\": cannot return classifier response"
                   << " because status is dirty" << std::endl;
      retval = 0;
   }
   else {
      if (IsNormalised()) {
         // normalise variables
         std::vector<double> iV;
         int ivar = 0;
         for (std::vector<double>::const_iterator varIt = inputValues.begin();
              varIt != inputValues.end(); varIt++, ivar++) {
            iV.push_back(NormVariable( *varIt, fVmin[ivar], fVmax[ivar] ));
         }
         retval = GetMvaValue__( iV );
      }
      else {
         retval = GetMvaValue__( inputValues );
      }
   }

   return retval;
}
